@@ -28,4 +28,4 @@ class dummyStart(Base):
 
     @transform_wrapper
     def transform(self, input):
-        return ""
+        return input
