@@ -30,3 +30,8 @@ def get_file_ext(filename):
                 ext = ""
         return ext
 
+def get_datetime_str(format=None):
+    today = datetime.now()
+    now_datetime_str = datetime.strftime(today, format)
+    return now_datetime_str
+
