@@ -104,7 +104,10 @@ def recursive_build_pipeline_from_cfg(  cfg,
 
             parallel_dict_index = [i for i in range(len(latest_nodes))]
             # print(parallel_dict_index, latest_nodes, trans_cfg)
-            for i, node_name, latest_node, sub_trans_cfg in zip(parallel_dict_index, pre_node_names, latest_nodes, trans_cfg["transformers"]):
+            for i, node_name, latest_node, sub_trans_cfg in zip(parallel_dict_index, 
+                                                                pre_node_names, 
+                                                                latest_nodes, 
+                                                                trans_cfg["transformers"]):
  
                 if "type" in sub_trans_cfg:
                     sub_trans_name = sub_trans_cfg.get("type")
@@ -140,16 +143,23 @@ def recursive_build_pipeline_from_cfg(  cfg,
 
 
                 elif "pipeline" in sub_trans_cfg:
-                    transformer, node_name_set, _node_pairs, node_name, _latest_node, index, parallel_index, nodename2transformer = recursive_build_pipeline_from_cfg( sub_trans_cfg, 
-                                                                                                                                                            node_name_set, 
-                                                                                                                                                            [], 
-                                                                                                                                                            node_name,
-                                                                                                                                                            latest_node, 
-                                                                                                                                                            index, 
-                                                                                                                                                            parallel_index, 
-                                                                                                                                                            nodename2transformer,
-                                                                                                                                                            pipeline_settings,
-                                                                                                                                                            kafka_media)
+                    transformer, \
+                    node_name_set, \
+                    _node_pairs, \
+                    node_name, \
+                    _latest_node, \
+                    index, \
+                    parallel_index, \
+                    nodename2transformer = recursive_build_pipeline_from_cfg(   sub_trans_cfg, 
+                                                                                node_name_set, 
+                                                                                [], 
+                                                                                node_name,
+                                                                                latest_node, 
+                                                                                index, 
+                                                                                parallel_index, 
+                                                                                nodename2transformer,
+                                                                                pipeline_settings,
+                                                                                kafka_media)
                     transformers_list.append(transformer)
                     parallel_node_pairs[f"process_no.{str(i)}"]=_node_pairs
                     parallel_latest_node.append(_latest_node)
@@ -177,16 +187,23 @@ def recursive_build_pipeline_from_cfg(  cfg,
             
         elif trans_name=="pipeline":
             # recursive function help update node_pairs, latest_node and index
-            _pipeline, node_name_set, _node_pairs, node_name, latest_node, index, parallel_index, nodename2transformer= recursive_build_pipeline_from_cfg( trans_cfg, 
-                                                                                                                                                node_name_set, 
-                                                                                                                                                [], 
-                                                                                                                                                node_name,
-                                                                                                                                                latest_node,
-                                                                                                                                                index, 
-                                                                                                                                                parallel_index, 
-                                                                                                                                                nodename2transformer,
-                                                                                                                                                pipeline_settings,
-                                                                                                                                                kafka_media)
+            _pipeline, \
+            node_name_set, \
+            _node_pairs, \
+            node_name, \
+            latest_node, \
+            index, \
+            parallel_index, \
+            nodename2transformer= recursive_build_pipeline_from_cfg(    trans_cfg, 
+                                                                        node_name_set, 
+                                                                        [], 
+                                                                        node_name,
+                                                                        latest_node,
+                                                                        index, 
+                                                                        parallel_index, 
+                                                                        nodename2transformer,
+                                                                        pipeline_settings,
+                                                                        kafka_media)
             # update steps
             node_pairs.append(_node_pairs)
             steps.append(_pipeline)
@@ -229,7 +246,14 @@ def recursive_build_pipeline_from_cfg(  cfg,
     node_name_set = set([node_name for node_name in node_name_set if node_name!=""])
 
 
-    return make_pipeline(*steps), node_name_set, node_pairs, node_name, latest_node, index, parallel_index, nodename2transformer
+    return make_pipeline(*steps), \
+            node_name_set, \
+            node_pairs, \
+            node_name, \
+            latest_node, \
+            index, \
+            parallel_index, \
+            nodename2transformer
             
 
 def create_tranformers_dict(transformers):
@@ -258,6 +282,7 @@ class Pipeline(sklearn_pipeline):
         self.pipe_start = pipe_start
         self.pipe_stop = pipe_stop
         self.delete_kafka_topic = 1
+        self.kafka_media=None
         
     
     def _islice(self, target_list, start, end):
@@ -342,6 +367,7 @@ class pipeline_settings():
         self.exchange_media = self.pipeline_settings["exchange_media"] if "exchange_media" in self.pipeline_settings else "default"
         self.delete_kafka_topic = self.pipeline_settings["delete_kafka_topic"] if "delete_kafka_topic" in self.pipeline_settings else 1
         self.bootstrap_servers = self.pipeline_settings["bootstrap_servers"] if "bootstrap_servers" in self.pipeline_settings else []
+        self.default_type = self.pipeline_settings["module_type"] if "default_type" in self.pipeline_settings else ""
     
     def get_pipeline_settings_from_cfg(self):
         
@@ -373,11 +399,15 @@ class DataPipeline:
             print(cfg)
             print("".join(["-" for i in range(80)]), "----------------------", "".join(["-" for i in range(80)]))
 
-        # self.kafka_media = kafka_media
-        from .kafka_media import kafkaMedia
-        self.kafka_media = kafkaMedia(bootstrap_servers=self._pipeline_settings.bootstrap_servers, topic="pipeline")
-        self.kafka_media.__enter__()
-        # print(self.kafka_media.producer, "#############################")
+
+        self.kafka_media = None
+        if self._pipeline_settings.exchange_media=="kafka":
+            if self._pipeline_settings.bootstrap_servers:
+                from .kafka_media import kafkaMedia
+                # print("self._pipeline_settings.bootstrap_servers", self._pipeline_settings.bootstrap_servers)
+                self.kafka_media = kafkaMedia(bootstrap_servers=self._pipeline_settings.bootstrap_servers, topic="pipeline")
+                self.kafka_media.__enter__()
+                # print(self.kafka_media.producer, "#############################")
 
 
         self._cfg = cfg
@@ -517,7 +547,16 @@ class DataPipeline:
         return self._pipeline.transform(input)
 
 
-def recursive_build_node_from_steps(steps, node_name_set, node_pairs, latest_node, index, parallel_index, nodename2transformer):
+def recursive_build_node_from_steps(steps, 
+                                    node_name_set, 
+                                    node_pairs, 
+                                    node_name,
+                                    latest_node, 
+                                    index, 
+                                    parallel_index, 
+                                    nodename2transformer,
+                                    pipeline_settings,
+                                    kafka_media):
 
     for step_tuple in steps:
 
@@ -539,15 +578,18 @@ def recursive_build_node_from_steps(steps, node_name_set, node_pairs, latest_nod
             # prepare latest_nodes
             if isinstance(latest_node, str):
                 latest_nodes = [latest_node for i in range(len(transformer.transformers))]
+                pre_node_names = [node_name for i in range(len(transformer.transformers))]
             elif isinstance(latest_node, list):
                 latest_nodes = latest_node
+                pre_node_names = node_name
             else:
                 assert False, "latest_node is not str or list"
 
             parallel_dict_index = [i for i in range(len(latest_nodes))]
 
             # dealing the transformers inside the parallelTransformer
-            for i, latest_node, sub_transformer in zip( parallel_dict_index, 
+            for i, node_name, latest_node, sub_transformer in zip( parallel_dict_index, 
+                                                        pre_node_names,
                                                         latest_nodes, 
                                                         transformer.transformers):
 
@@ -555,6 +597,7 @@ def recursive_build_node_from_steps(steps, node_name_set, node_pairs, latest_nod
                 if isinstance(sub_transformer, DataPipeline):
                     node_name_set, \
                     _node_pairs, \
+                    node_name, \
                     latest_node, \
                     index, \
                     parallel_index, \
@@ -564,17 +607,23 @@ def recursive_build_node_from_steps(steps, node_name_set, node_pairs, latest_nod
                                                                             latest_node, 
                                                                             index, 
                                                                             parallel_index, 
-                                                                            nodename2transformer)
+                                                                            nodename2transformer,
+                                                                            pipeline_settings,
+                                                                            kafka_media)
                     node_pairs.append(_node_pairs)
 
                 else:
                 ####################################### dealing with real transformer
                     index = index + 1
                     description = sub_transformer.description if sub_transformer.description else ""
+                    sub_transformer.pre_node_name = node_name
                     sub_transformer.index = index
                     _tuple = _name_estimators([sub_transformer])
                     sub_transformer.node_name = f"{str(index)}.{_tuple[0][0]}"
                     sub_transformer.description = description
+                    sub_transformer.print_task_result = pipeline_settings.print_task_result
+                    sub_transformer.exchange_media = pipeline_settings.exchange_media
+                    sub_transformer.kafka_media = kafka_media
                     transformers_list.append(sub_transformer)
                     # update nodename2transformer
                     nodename2transformer[sub_transformer.node_name] = sub_transformer
@@ -596,27 +645,37 @@ def recursive_build_node_from_steps(steps, node_name_set, node_pairs, latest_nod
         elif isinstance(transformer, DataPipeline):
             node_name_set, \
             _node_pairs, \
+            node_name, \
             latest_node, \
             index, \
             parallel_index, \
             nodename2transformer = recursive_build_node_from_steps( sub_transformer.steps,
                                                                     node_name_set,
-                                                                    node_pairs, 
+                                                                    [], 
                                                                     latest_node, 
                                                                     index, 
                                                                     parallel_index, 
-                                                                    nodename2transformer)
+                                                                    nodename2transformer,
+                                                                    pipeline_settings,
+                                                                    kafka_media)
             node_pairs.append(_node_pairs)
         
         else:
 
             index = index + 1
+            pre_node_name = node_name
             description = transformer.description if transformer.description else ""
             transformer.index = index
             transformer.node_name = f"{str(index)}.{trans_name}"
             transformer.description = description
+            transformer.print_task_result = pipeline_settings.print_task_result
+            transformer.exchange_media = pipeline_settings.exchange_media
+            transformer.kafka_media = kafka_media
+
             # update nodename2transformer
             nodename2transformer[transformer.node_name] = transformer
+            # set pre_node_name
+            transformer.pre_node_name = pre_node_name
             # update node_pairs
             if isinstance(latest_node, str):
                 node_pairs.append((latest_node, transformer.node_name))
@@ -650,10 +709,13 @@ def build_pipeline(*steps):
     self._nodename2transformer = recursive_build_node_from_steps(   steps_tuples,
                                                                     self._node_name_set,
                                                                     self._node_pairs, 
+                                                                    "",
                                                                     self._latest_node, 
                                                                     self._index,
                                                                     self._parallel_index,
-                                                                    self._nodename2transformer)
+                                                                    self._nodename2transformer,
+                                                                    self._pipeline_settings,
+                                                                    self.kafka_media)
 
     # self.nodes_text = init_nodes_text(self._nodename2transformer)
 
